#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          15-05-2018
//      Created for Project :   FLAME
"""

from pyramid.view import view_defaults
from pyramid.httpexceptions import HTTPBadRequest, HTTPInternalServerError
from influxdb import InfluxDBClient
from urllib.parse import urlparse
from subprocess import Popen
from clmcservice.aggregationapi.utilities import validate_config_content, validate_action_content, validate_round_trip_query_params, \
    CONF_OBJECT, CONF_FILE_ATTRIBUTE, AGGREGATOR_CONFIG_SECTION, CONFIG_ATTRIBUTES, ROUND_TRIP_ATTRIBUTES, RUNNING_FLAG, PROCESS_ATTRIBUTE, MALFORMED_FLAG, COMMENT_ATTRIBUTE, COMMENT_VALUE
import os
import os.path
import sys
import logging
import configparser


log = logging.getLogger('service_logger')


@view_defaults(route_name='aggregator_config', renderer='json')
class AggregatorConfig(object):
    """
    A class-based view for accessing and mutating the configuration of the aggregator.
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    def get(self):
        """
        A GET API call for the configuration of the aggregator.

        :return: A JSON response with the configuration of the aggregator.
        """

        aggregator_config_data = self.request.registry.settings[CONF_OBJECT]  # fetch the configuration object
        if aggregator_config_data is None:
            raise HTTPBadRequest("Aggregator has not been configured, yet. Send a PUT request to /aggregator/config with a JSON body of the configuration.")

        config = {key: aggregator_config_data[AGGREGATOR_CONFIG_SECTION][key] for key in CONFIG_ATTRIBUTES}  # extract a json value containing the config attributes
        config['aggregator_report_period'] = int(config['aggregator_report_period'])

        return config

    def put(self):
        """
        A PUT API call for the status of the aggregator.

        :return: A JSON response to the PUT call - essentially with the new configured data and comment of the state of the aggregator
        :raises HTTPBadRequest: if request body is not a valid JSON for the configurator
        """

        try:
            new_config = self.request.body.decode(self.request.charset)
            new_config = validate_config_content(new_config)  # validate the content and receive a json dictionary object
        except AssertionError as e:
            raise HTTPBadRequest("Bad request content. Configuration format is incorrect: {0}".format(e.args))

        conf = self.request.registry.settings[CONF_OBJECT]
        if conf is None:
            conf = configparser.ConfigParser()
            conf[AGGREGATOR_CONFIG_SECTION] = {}
            self.request.registry.settings[CONF_OBJECT] = conf
            old_config = {}
        else:
            # save the old configuration before updating so that it can be compared to the new one and checked for malformed state
            old_config = {attribute: conf[AGGREGATOR_CONFIG_SECTION][attribute] for attribute in CONFIG_ATTRIBUTES}
            old_config['aggregator_report_period'] = int(old_config['aggregator_report_period'])

        for attribute in CONFIG_ATTRIBUTES:
            conf[AGGREGATOR_CONFIG_SECTION][attribute] = str(new_config.get(attribute))  # update the configuration attributes

        # if configuration is not already malformed, check whether the configuration is updated (changed in any way), if so (and the aggregator is running), malformed state is detected
        if not self.request.registry.settings[MALFORMED_FLAG]:
            malformed = old_config != new_config and AggregatorController.is_process_running(self.request.registry.settings.get(PROCESS_ATTRIBUTE))
            self.request.registry.settings[MALFORMED_FLAG] = malformed
            if malformed:
                new_config[MALFORMED_FLAG] = True
                new_config[COMMENT_ATTRIBUTE] = COMMENT_VALUE

        self._write_conf_file()  # save the updated configuration to conf file
        return new_config

    def _write_conf_file(self):
        """
        Writes the configuration settings of the aggregator to a file with path stored at CONF_FILE_ATTRIBUTE
        """

        conf = self.request.registry.settings[CONF_OBJECT]
        conf_file_path = self.request.registry.settings[CONF_FILE_ATTRIBUTE]
        os.makedirs(os.path.dirname(conf_file_path), exist_ok=True)

        log.info("Saving configuration to file {0}.".format(conf_file_path))
        with open(conf_file_path, 'w') as configfile:
            log.info("Opened configuration file {0}.".format(conf_file_path))
            conf.write(configfile)
        log.info("Successfully saved configuration to file {0}.".format(conf_file_path))


@view_defaults(route_name='aggregator_controller', renderer='json')
class AggregatorController(object):

    """
    A class-based view for controlling the aggregator.
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    def get(self):
        """
        A GET API call for the status of the aggregator - running or not.

        :return: A JSON response with the status of the aggregator.
        """

        aggregator_data = self.request.registry.settings
        aggregator_process = aggregator_data.get(PROCESS_ATTRIBUTE)
        aggregator_running = self.is_process_running(aggregator_process)

        config = {RUNNING_FLAG: aggregator_running}

        if aggregator_data[MALFORMED_FLAG] and aggregator_running:
            config[MALFORMED_FLAG] = True
            config[COMMENT_ATTRIBUTE] = COMMENT_VALUE

        return config

    def put(self):
        """
        A PUT API call for the status of the aggregator.

        :return: A JSON response to the PUT call - essentially saying whether the aggregator is running or not
        :raises HTTPBadRequest: if request body is not a valid JSON for the controller
        """

        content = self.request.body.decode(self.request.charset)

        try:
            content = validate_action_content(content)

            conf = self.request.registry.settings[CONF_OBJECT]
            if conf is None:
                raise HTTPBadRequest("You must configure the aggregator before controlling it. Send a PUT request to /aggregator/config with a JSON body of the configuration.")

            aggregator_config = {attribute: conf[AGGREGATOR_CONFIG_SECTION][attribute] for attribute in CONFIG_ATTRIBUTES}
            aggregator_config['aggregator_report_period'] = int(aggregator_config['aggregator_report_period'])

            action = content['action']

            aggregator_running = self.is_process_running(self.request.registry.settings.get(PROCESS_ATTRIBUTE))
            if action == 'start':
                if not aggregator_running:
                    process = self.start_aggregator(aggregator_config)
                    aggregator_running = True
                    self.request.registry.settings[PROCESS_ATTRIBUTE] = process
                    self.request.registry.settings[MALFORMED_FLAG] = False
            elif action == 'stop':
                self.stop_aggregator(self.request.registry.settings.get(PROCESS_ATTRIBUTE))
                aggregator_running = False
                self.request.registry.settings[PROCESS_ATTRIBUTE] = None
                self.request.registry.settings[MALFORMED_FLAG] = False
            elif action == 'restart':
                self.stop_aggregator(self.request.registry.settings.get(PROCESS_ATTRIBUTE))
                process = self.start_aggregator(aggregator_config)
                aggregator_running = True
                self.request.registry.settings[PROCESS_ATTRIBUTE] = process
                self.request.registry.settings[MALFORMED_FLAG] = False

            return {RUNNING_FLAG: aggregator_running}

        except AssertionError:
            raise HTTPBadRequest('Bad request content - must be in JSON format: {"action": value}, where value is "start", "stop" or "restart".')

    @staticmethod
    def start_aggregator(config):
        """
        An auxiliary method to start the aggregator.

        :param config: the configuration containing the arguments for the aggregator
        :return: the process object of the started aggregator script
        """

        python_interpreter = sys.executable
        command = [python_interpreter, '-m', 'clmcservice.aggregation.aggregator', '--period', str(config.get('aggregator_report_period')), '--database',
                   config.get('aggregator_database_name'), '--url', config.get('aggregator_database_url')]
        process = Popen(command)

        log.info("\nStarted aggregator process with PID: {0}\n".format(process.pid))

        return process

    @staticmethod
    def stop_aggregator(process):
        """
        An auxiliary method to stop the aggregator.

        :param process: the process to terminate
        """

        # check if the process is started
        if AggregatorController.is_process_running(process):
            process.terminate()
            log.info("\nStopped aggregator process with PID: {0}\n".format(process.pid))

    @staticmethod
    def is_process_running(process):
        """
        Checks if a process is running.

        :param process: the Popen object to check
        :return: True if running, False otherwise
        """

        # check if the process is started before trying to terminate it - process.poll() only returns something if the process has terminated, hence we check for a None value
        return process is not None and process.poll() is None


@view_defaults(route_name='round_trip_time_query', renderer='json')
class RoundTripTimeQuery(object):

    # TODO This API endpoint has not been tested, neither has the formula used to calculate round trip time.

    """
    A class-based view for querying the round trip time in a given range.
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    def get(self):
        """
        A GET API call for the averaged round trip time of a specific media service over a given time range.

        :return: A JSON response with the round trip time and its contributing parts.
        """

        params = {}
        for attribute in ROUND_TRIP_ATTRIBUTES:
            if attribute in self.request.params:
                params[attribute] = self.request.params.get(attribute)

        try:
            params = validate_round_trip_query_params(params)

            conf = self.request.registry.settings[CONF_OBJECT]
            if conf is None:
                raise HTTPBadRequest("You must configure the aggregator before making a round trip time query. Send a PUT request to /aggregator/config with a JSON body of the configuration.")

            aggregator_config_data = {config_attribute: conf[AGGREGATOR_CONFIG_SECTION][config_attribute] for config_attribute in CONFIG_ATTRIBUTES}
            aggregator_config_data['aggregator_report_period'] = int(aggregator_config_data['aggregator_report_period'])

            media_service = params.get(ROUND_TRIP_ATTRIBUTES[0])
            start_timestamp = params.get(ROUND_TRIP_ATTRIBUTES[1])
            end_timestamp = params.get(ROUND_TRIP_ATTRIBUTES[2])
            influx_db_name = aggregator_config_data.get(CONFIG_ATTRIBUTES[1])
            influx_db_url = aggregator_config_data.get(CONFIG_ATTRIBUTES[2])

            url_object = urlparse(influx_db_url)
            try:
                db_client = InfluxDBClient(host=url_object.hostname, port=url_object.port, database=influx_db_name, timeout=10)
                query = 'SELECT mean(*) FROM "{0}"."autogen"."e2e_delays" WHERE time >= {1} and time < {2} and sf_instance = \'{3}\''.format(
                    influx_db_name, start_timestamp, end_timestamp, media_service)
                log.info("Executing query: {0}".format(query))
                result = db_client.query(query)

                actual_result = next(result.get_points(), None)
                if actual_result is None:
                    return {"result": None}
                else:
                    forward_latency = actual_result.get("mean_delay_forward")
                    reverse_latency = actual_result.get("mean_delay_reverse")
                    service_delay = actual_result.get("mean_delay_service")
                    request_size = actual_result.get("mean_avg_request_size")
                    response_size = actual_result.get("mean_avg_response_size")
                    bandwidth = actual_result.get("mean_avg_bandwidth")

                    rtt = self.calculate_round_trip_time(forward_latency, reverse_latency, service_delay, request_size, response_size, bandwidth)
                    return {"result": rtt}
            except Exception as e:
                msg = "Cannot instantiate connection with database {0} on url {1}.".format(influx_db_name, influx_db_url)
                log.info(msg)
                log.error(type(e))
                log.error(e)
                log.error(e.args)

                raise HTTPInternalServerError(msg)

        except AssertionError:
            raise HTTPBadRequest('Bad request content - must be in JSON format: {"media_service": value, "start_timestamp": value, "end_timestamp": value}.')

    @staticmethod
    def calculate_round_trip_time(forward_latency, reverse_latency, service_delay, request_size, response_size, bandwidth, packet_size=1500, packet_header_size=50):
        """
        Calculates the round trip time given the list of arguments.

        :param forward_latency: network latency in forward direction (s)
        :param reverse_latency: network latency in reverse direction (s)
        :param service_delay: media service delay (s)
        :param request_size: request size (bytes)
        :param response_size: response size (bytes)
        :param bandwidth: network bandwidth (Mb/s)
        :param packet_size: size of packet (bytes)
        :param packet_header_size: size of the header of the packet (bytes)
        :return: the calculated round trip time
        """

        forward_data_delay = (8/10**6) * (request_size / bandwidth) * (packet_size / (packet_size - packet_header_size))
        reverse_data_delay = (8/10**6) * (response_size / bandwidth) * (packet_size / (packet_size - packet_header_size))

        return forward_latency + forward_data_delay + service_delay + reverse_latency + reverse_data_delay
